﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace SqlDependencyRefresher
{
    public partial class MainForm : Form
    {
        public enum DependencyType
        {
            Unknown,
            View,
            Function,
            Procedure,
            Trigger
        }

        private class DependentObject
        {
            public string Name { get; private set; }
            public DependencyType Type { get; private set; }

            public DependentObject(string name, DependencyType type)
            {
                Name = name;
                Type = type;
            }
        }

        private DataTable _dataTable;

        public MainForm()
        {
            InitializeComponent();
            btnGenerate.Enabled = txtFilter.Enabled = dgvTables.Enabled = false;

            BackgroundWorker worker = new BackgroundWorker();
            worker.DoWork += (s, e) =>
                {
                    e.Result = TestConnection();
                };
            worker.RunWorkerCompleted += (s, e) =>
                {
                    lblPopulating.Visible = false;
                    if ((bool)e.Result)
                    {
                        InitializeDataTable();
                        PopulateGrid();
                        btnGenerate.Enabled = txtFilter.Enabled = dgvTables.Enabled = true;
                    }
                    else
                    {
                        MessageBox.Show("SQL connection error. \nPlease configure the App.Config with correct connection settings.");
                    }
                };
            worker.RunWorkerAsync();
        }

        private void InitializeDataTable()
        {
            _dataTable = new DataTable();
            _dataTable.Columns.Add("Name", typeof(string));
            _dataTable.Columns.Add("Refresh", typeof(bool));
        }

        private bool TestConnection()
        {
            try
            {
                using (SqlConnection connection = new SqlConnection(GetConnectionString()))
                {
                    connection.Open();
                }
            }
            catch (SqlException)
            {
                return false;
            }
            return true;
        }

        private void PopulateGrid()
        {
            _dataTable.Rows.Clear();

            using (SqlConnection connection = new SqlConnection(GetConnectionString()))
            {
                string cmd = "SELECT TABLE_SCHEMA, TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='BASE TABLE' ORDER BY TABLE_NAME";

                SqlCommand command = new SqlCommand(cmd, connection);
                connection.Open();

                using (SqlDataReader reader = command.ExecuteReader())
                {
                    while (reader.Read())
                    {
                        string tableName = String.Format("[{0}].[{1}]", reader.GetString(0), reader.GetString(1));
                        _dataTable.Rows.Add(tableName, false);
                    }
                }
            }

            dgvTables.AutoGenerateColumns = false;
            dgvTables.Columns[0].DataPropertyName = "Name";
            dgvTables.Columns[1].DataPropertyName = "Refresh";
            dgvTables.DataSource = _dataTable;
        }

        private string GetConnectionString()
        {
            return ConfigurationManager.ConnectionStrings["Default"].ConnectionString;
        }

        private void Generate_Click(object sender, EventArgs e)
        {
            StringBuilder sb = new StringBuilder();
            foreach (DataRow row in _dataTable.Rows)
            {
                bool refresh = (bool)row["Refresh"];
                if (refresh)
                {
                    string name = row["Name"].ToString();
                    foreach (var dependent in GetDependantObjects(name))
                    {
                        sb.AppendLine();
                        sb.AppendLine(GenerateDropStatement(dependent));
                        sb.AppendLine();
                        sb.AppendLine(GetObjectSqlText(dependent));
                        sb.AppendLine();
                        sb.AppendLine("GO");
                        sb.AppendLine();
                    }
                }
            }

            var fileName = "Temp_GenQuery.txt";
            File.WriteAllText(fileName, sb.ToString());
            Process.Start(fileName);
        }

        private string GenerateDropStatement(DependentObject obj)
        {
            StringBuilder sb = new StringBuilder();
            switch (obj.Type)
            {
                case (DependencyType.View):
                    sb.AppendLine(String.Format("IF  EXISTS (SELECT * FROM sys.views WHERE object_id = OBJECT_ID(N'{0}'))", obj.Name));
                    sb.AppendLine(String.Format("DROP VIEW {0}", obj.Name));
                    break;

                case (DependencyType.Procedure):
                    sb.AppendLine(String.Format("IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'{0}') AND type in (N'P', N'PC'))", obj.Name));
                    sb.AppendLine(String.Format("DROP PROCEDURE {0}", obj.Name));
                    break;

                case (DependencyType.Function):
                    sb.AppendLine(String.Format("IF  EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'{0}') AND type in (N'FN', N'IF', N'TF', N'FS', N'FT'))", obj.Name));
                    sb.AppendLine(String.Format("DROP FUNCTION {0}", obj.Name));
                    break;

                case (DependencyType.Trigger):
                    sb.AppendLine(String.Format("IF  EXISTS (SELECT * FROM sys.triggers WHERE object_id = OBJECT_ID(N'{0}'))", obj.Name));
                    sb.AppendLine(String.Format("DROP TRIGGER {0}", obj.Name));
                    break;

                default:
                    sb.AppendLine(String.Format("PRINT 'COULD NOT GENERATE DROP FUNCTION FOR {0}'", obj.Name));
                    break;
            }

            sb.AppendLine("GO");

            return sb.ToString();
        }

        private List<DependentObject> GetDependantObjects(string name)
        {
            List<DependentObject> dependentObjects = new List<DependentObject>();
            using (SqlConnection conn = new SqlConnection(GetConnectionString()))
            {
                string cmd = String.Format("EXEC sp_depends '{0}'", name);

                DataSet dataset = new DataSet();

                SqlDataAdapter adapter = new SqlDataAdapter();
                adapter.SelectCommand = new SqlCommand(cmd, conn);
                adapter.Fill(dataset);

                foreach (DataTable table in dataset.Tables)
                {
                    foreach (DataRow row in table.Rows)
                    {
                        var objName = (string)row.ItemArray[0];
                        var type = GetDependencyType((string)row.ItemArray[1]);

                        dependentObjects.Add(new DependentObject(objName, type));
                    }
                }
            }
            return dependentObjects;
        }

        private DependencyType GetDependencyType(string s)
        {
            s = s.ToLower();
            if (s.Contains("view")) return DependencyType.View;
            if (s.Contains("function")) return DependencyType.Function;
            if (s.Contains("procedure")) return DependencyType.Procedure;
            if (s.Contains("trigger")) return DependencyType.Trigger;
            return DependencyType.Unknown;
        }

        private string GetObjectSqlText(DependentObject obj)
        {
            if (obj.Type == DependencyType.Unknown)
            {
                return String.Format("PRINT 'COULD NOT GENERATE CREATE FUNCTION FOR {0}'", obj.Name);
            }

            using (SqlConnection connection = new SqlConnection(GetConnectionString()))
            {
                string cmd = String.Format("EXEC sp_helptext '{0}'", obj.Name);

                SqlCommand command = new SqlCommand(cmd, connection);
                connection.Open();

                using (SqlDataReader reader = command.ExecuteReader())
                {
                    StringBuilder sb = new StringBuilder();
                    while (reader.Read())
                    {
                        sb.Append(reader.GetString(0));
                    }
                    return sb.ToString();
                }
            }
        }

        private void Filter_TextChanged(object sender, EventArgs e)
        {
            _dataTable.DefaultView.RowFilter = string.Format("Name LIKE '%{0}%'", txtFilter.Text);
        }
    }
}